/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WLINESEARCH_H
#define WLINESEARCH_H

#include "tbox.h"
#include "wObject.h"
#include <iostream>

#ifndef SWIG

namespace tbox
{

/**
 * @brief Linesearch function
 * @authors Romain Boman
 */
class TBOX_API LSFunction : public fwk::wObject
{
public:
    LSFunction();
    virtual double eval(double alpha);
};

class TBOX_API RevFunction : public LSFunction
{
    LSFunction &f;

public:
    RevFunction(LSFunction &_f);
    virtual double eval(double alpha) override;
};

/**
 * @brief Linesearch base class
 * @authors Adrien Crovato
 */
class TBOX_API Linesearch
{
protected:
    LSFunction &fct; ///< function
    size_t maxLsIt;  ///< max. number of iterations for linesearch
    double tol;      ///< tolerance for linesearch
    int verbose;     ///< display more info

public:
    size_t fevalIt; ///< function evaluation count

    Linesearch(LSFunction &_fct);

    virtual void set(int _maxLSIt, double _lSTol, bool _verbose);
    virtual double run();
};

/**
 * @brief Simple backtrack linesearch
 * @authors Adrien Crovato
 */
class TBOX_API BacktrackLS : public Linesearch
{
public:
    BacktrackLS(LSFunction &_fct);

    virtual double run() override;
};

/**
 * @brief 3-points quadratic linesearch
 *
 * Implemented following Prof. Fleury notes, University of Liege
 * @authors Romain Boman, Adrien Crovato
 */
class TBOX_API FleuryLS : public Linesearch
{
private:
    double hU;       ///< upper bound value for bracketing
    double hS;       ///< current value for bracketing
    double h0;       ///< initial value for bracketing
    double hL;       ///< lower bound value for bracketing
    bool revAllowed; ///< allow bracketing in reverse direction

public:
    size_t brckIt; ///< bracketing iteration count
    size_t lsIt;   ///< linesearch iteration count

    FleuryLS(LSFunction &_fct, double _h0 = 1, double _hL = 1e-10, bool _revAllowed = true);

    virtual double run() override;
};

/**
 * @brief Bank and Rose linesearch
 *
 * Implemented following Bank and Rose, Global Approximate Newton Methods, Numerische Mathematik
 * www.math.hcmuns.edu.vn/~tangoc//bbmot.pdf
 * @authors Adrien Crovato
 */
class TBOX_API BankRoseLS : public Linesearch
{
private:
    double K; ///> damping control

public:
    BankRoseLS(LSFunction &_fct);
    virtual double run() override;
};

} // namespace tbox

#endif

#endif // WLINESEARCH_H
