/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WMSHCRACK_H
#define WMSHCRACK_H

#include "tbox.h"
#include "wObject.h"
#include <Eigen/Dense>
#include <vector>
#include <map>
#include <unordered_set>
#include <memory>

namespace tbox
{

/**
 * @brief Duplicate nodes and elements in a physical group and make mesh consistent
 *
 * Similar to plugin "crack" from gmsh but allows more control. Currently implemented
 * only for Line2 and Tri3.
 * @authors Adrien Crovato
 * @todo Automatic identification of boundary surface elements on negative side of crack is not robust (might fail for concave boundaries)
 */
class TBOX_API MshCrack : public fwk::wSharedObject
{
private:
    std::shared_ptr<MshData> msh;    ///< mesh
    int nDim;                        ///< number of dimensions
    bool hasCrack;                   ///< flag telling if mesh already has crack
    std::map<Node *, Node *> nodMap; ///< map between nodes on both side of crack
    Group *crckGrp;                  ///< physical group of crack
    Group *exclGrp;                  ///< physical group of nodes not to be duplicated
    std::vector<Groups *> sBndGrps;  ///< list of pair of surface physical groups touching crack
    std::vector<Groups *> vBndGrps;  ///< list of pair of volume physical groups touching crack
    std::vector<Group *> sBndGrp;    ///< list of surface physical groups touching crack
    std::vector<Group *> vBndGrp;    ///< list of volume physical groups touching crack

    void openCrack();
    void modifyBoundaries();
    std::unordered_set<Element *> findTouching(std::vector<Element *> const &elms, std::unordered_set<Node *> const &nods);
    size_t findClosest(std::vector<Eigen::Vector3d> const &cga, Eigen::Vector3d const &cgb);
    void mergeTags(std::vector<Groups *> &grps);
    void swapNodes(std::vector<Element *> &elms);
    void writeElems(std::vector<int> const &elmsNo, std::string const filename) const;

public:
    MshCrack(std::shared_ptr<MshData> _msh, int _nDim);
    virtual ~MshCrack();

    void setCrack(std::string const &name);
    void setExcluded(std::string const &name);
    void addBoundary(std::string const &name);
    void run();

#ifndef SWIG
    virtual void write(std::ostream &out) const override;
#endif
};

} // namespace tbox

#endif // WMSHCRACK_H