#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @brief VTK cut plane post-processing
# @authors Adrien Crovato, Kim Liegeois

try:
    import vtk
except:
    raise Exception('VTK not found. Aborting!\n')

class LineExtractor:
    def __init__(self, _grid):
        self.grid = _grid

    def cut(self, tag, p1, p2, tag_name="tag"):
        '''
        Cut the physical group ided tag_name="tag" with a line defined by point "p1" and normal "p2"
        '''
        import numpy as np
        self.p1 = np.zeros((3,))
        self.p2 = np.zeros((3,))
        for j in range(0, 3):
          self.p1[j] = p1[j]
          self.p2[j] = p2[j]
        
        p12 = self.p2-self.p1
        
        v1 = np.zeros((3,))
        v1[0] = 1
        
        if np.linalg.norm(np.cross(v1,p12)) != 0:
          v1 = np.cross(v1, p12)
          v2 = np.cross(v1, p12)
        else:
          v1[0] = 0
          v1[1] = 1
          v2 = np.zeros((3,))
          v2[2] = 1

        # create a threshold containing the physical group (tag) to cut
        thresh = vtk.vtkThreshold()
        thresh.SetLowerThreshold(tag)
        thresh.SetUpperThreshold(tag)
        thresh.SetInputConnection(self.grid)
        thresh.SetInputArrayToProcess(0, 0, 0, vtk.vtkDataObject.FIELD_ASSOCIATION_CELLS, tag_name) # "tag" is the default name set in tboxVtk::VtkExport::save
        thresh.Update()
        # create the first cutting plane
        plane1 = vtk.vtkPlane()
        plane1.SetOrigin(p1[0], p1[1], p1[2])
        plane1.SetNormal(v1[0], v1[1], v1[2])
        plane2 = vtk.vtkPlane()
        plane2.SetOrigin(p1[0], p1[1], p1[2])
        plane2.SetNormal(v2[0], v2[1], v2[2]) 
        # cut the threshold and get data
        cutter1 = vtk.vtkFiltersCorePython.vtkCutter()
        cutter1.SetCutFunction(plane1)
        cutter1.SetInputConnection(thresh.GetOutputPort())
        cutter1.Update()

        cutter2 = vtk.vtkFiltersCorePython.vtkCutter()
        cutter2.SetCutFunction(plane2)
        cutter2.SetInputConnection(cutter1.GetOutputPort())
        cutter2.Update()

        return cutter2.GetOutput()

    def extract(self, cutData, vname, atPoint = True):
        '''
        Extract points "pts", connectivity list "elems" and data "vals" named "vname" from cutting plane data "cutData".
        The physical group dimension is given by "tagDim" and "atPoint" is True if vname is defined at points.
        '''
        import numpy as np
        # transfer point coordinates
        norm_p1p2 = np.linalg.norm(self.p2-self.p1)
        _pts = cutData.GetPoints()
        pts = np.zeros((_pts.GetNumberOfPoints(), ))
        for i in range(0, len(pts)):
            tmp = np.zeros((3,))
            for j in range(0, 3):
                tmp[j] = _pts.GetPoint(i)[j]
            pts[i] = np.inner(self.p2-self.p1,tmp-self.p1) / norm_p1p2
        indices_tmp = np.argsort(pts)
        pts = pts
        pts_res = pts[indices_tmp]
        indices_2 = np.argwhere((pts_res >= 0) & (pts_res <= norm_p1p2))
        pts_res = pts_res[indices_2]
        # transfer connectivity
        _elems = cutData.GetLines().GetData()
        nV = 2
        max_nb_elems = _elems.GetNumberOfTuples() // (nV+1)
        current_nb_elems = 0
        indices_elems = np.zeros((max_nb_elems,), dtype=int)
        for i in range(0, max_nb_elems):
            has_to_be_kept = True
            for j in range(0, nV):
                node_id = int(_elems.GetTuple((nV+1)*i+j+1)[0])
                if pts[node_id] < 0 or pts[node_id] > norm_p1p2:
                    has_to_be_kept = False
                    break
            if has_to_be_kept:
                current_nb_elems = current_nb_elems + 1
                indices_elems[i] = 1
        indices_elems_2 = np.zeros((current_nb_elems,), dtype=int)
        elems = np.zeros((current_nb_elems, nV), dtype=int)
        mean_per_elems = np.zeros((current_nb_elems,))
        current_i = -1
        for i in range(0, max_nb_elems):
            if indices_elems[i]:
                current_i = current_i + 1
                indices_elems_2[current_i] = i
                for j in range(0, nV):
                    node_id = int(_elems.GetTuple((nV+1)*i+j+1)[0])
                    elems[current_i][j] = np.argwhere(pts[node_id] == pts_res)[0][0]
                    mean_per_elems[current_i] = mean_per_elems[current_i] + 0.5*pts[node_id]
        indices_elems_3 = np.argsort(mean_per_elems)
        elems = elems[indices_elems_3][:]
        # transfer variables
        vals = {}
        for name in vname:
            if atPoint: # data at points
                _vals =  cutData.GetPointData().GetArray(name)
            else: # data at elements
                _vals =  cutData.GetCellData().GetArray(name)
            vals[name] = np.zeros((_vals.GetNumberOfTuples(), _vals.GetNumberOfComponents()))
            for i in range(0, vals[name].shape[0]):
                for j in range(0, vals[name].shape[1]):
                    if atPoint:
                        vals[name][i][j] = _vals.GetTuple(i)[j]
                    else:
                        vals[name][i][j] = _vals.GetTuple(i)[j]
            if atPoint:
                tmp = vals[name][indices_tmp][:]
                vals[name] = tmp[indices_2][:]
            else:
                tmp = vals[name][indices_elems_2][:]
                vals[name] = tmp[indices_elems_3][:]

        return pts_res, elems, vals
