/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wVtkExport.h"
#include "wNode.h"
#include "wElement.h"
#include "wResults.h"
#include "wTag.h"

#include <vtkSmartPointer.h>
#include <vtkUnstructuredGrid.h>
#include <vtkXMLUnstructuredGridWriter.h>
#include <vtkPoints.h>
#include <vtkHexahedron.h>
#include <vtkTetra.h>
#include <vtkQuad.h>
#include <vtkTriangle.h>
#include <vtkLine.h>
#include <vtkVertex.h>
#include <vtkDoubleArray.h>
#include <vtkIntArray.h>
#include <vtkPointData.h>
#include <vtkCellData.h>

using namespace tbox;
using namespace tboxVtk;

using vtkWriterP = vtkSmartPointer<vtkXMLUnstructuredGridWriter>;
using vtkPtsP = vtkSmartPointer<vtkPoints>;
using vtkHexaP = vtkSmartPointer<vtkHexahedron>;
using vtkTetraP = vtkSmartPointer<vtkTetra>;
using vtkQuadP = vtkSmartPointer<vtkQuad>;
using vtkTriP = vtkSmartPointer<vtkTriangle>;
using vtkLineP = vtkSmartPointer<vtkLine>;
using vtkVtxP = vtkSmartPointer<vtkVertex>;
using vtkDblArrP = vtkSmartPointer<vtkDoubleArray>;
using vtkIntArrP = vtkSmartPointer<vtkIntArray>;

VtkExport::VtkExport(std::shared_ptr<MshData> _msh) : MshExport(_msh)
{
    grid = vtkSmartPointer<vtkUnstructuredGrid>::New();
    create();
}

VtkExport::~VtkExport()
{
    std::cout << "~VtkExport()\n";
}

/**
 * @brief Save mesh on disk
 * @authors Adrien Crovato
 */
void VtkExport::save(std::string const &fname) const
{
    // Update point position
    grid->GetPoints()->Reset();
    for (auto n : msh->nodes)
        grid->GetPoints()->InsertPoint(n->no - 1, n->pos(0), n->pos(1), n->pos(2));
    grid->GetPoints()->Modified();

    // Define writer
    std::string wname = fname + ".vtu";
    vtkWriterP writer = vtkWriterP::New();
    writer->SetFileName(wname.c_str());
    writer->SetInputData(grid);
    // Write binary or ascii
    if (binary)
    {
        std::cout << "writing binary file: " << wname << "... " << std::flush;
        writer->SetDataModeToBinary();
    }
    else
    {
        std::cout << "writing ascii file: " << wname << "... " << std::flush;
        writer->SetDataModeToAscii();
    }
    writer->Write();
    std::cout << "done!" << std::endl;
}

/**
 * @brief Save mesh and results on disk
 * @authors Adrien Crovato, Kim Liegeois
 */
void VtkExport::save(std::string const &fname, Results const &r) const
{
    // Create scalar field (at nodes)
    for (auto &p : r.scalars_at_nodes)
    {
        vtkDblArrP data = vtkDblArrP::New();
        data->SetNumberOfComponents(1);
        data->SetNumberOfTuples(msh->nodes.size());
        data->SetName(p.first.c_str());
        for (size_t i = 0; i < msh->nodes.size(); ++i)
            data->SetValue(msh->nodes[i]->no - 1, (*(p.second))[i]);
        grid->GetPointData()->AddArray(data);
    }
    // Create vector field (at nodes)
    for (auto &p : r.vectors_at_nodes)
    {
        vtkDblArrP data = vtkDblArrP::New();
        data->SetNumberOfComponents(3);
        data->SetNumberOfTuples(msh->nodes.size());
        data->SetName(p.first.c_str());
        for (size_t i = 0; i < msh->nodes.size(); ++i)
            data->SetTuple3(msh->nodes[i]->no - 1, (*(p.second))[i](0), (*(p.second))[i](1), (*(p.second))[i](2));
        grid->GetPointData()->AddArray(data);
    }
    // Create tensor field (at nodes)
    for (auto &p : r.tensors_at_nodes)
    {
        vtkDblArrP data = vtkDblArrP::New();
        data->SetNumberOfComponents(9);
        data->SetNumberOfTuples(msh->nodes.size());
        data->SetName(p.first.c_str());
        for (size_t i = 0; i < msh->nodes.size(); ++i)
            data->SetTuple9(msh->nodes[i]->no - 1, (*(p.second))[i](0, 0), (*(p.second))[i](0, 1), (*(p.second))[i](0, 2),
                            (*(p.second))[i](1, 0), (*(p.second))[i](1, 1), (*(p.second))[i](1, 2),
                            (*(p.second))[i](2, 0), (*(p.second))[i](2, 1), (*(p.second))[i](2, 2));
        grid->GetPointData()->AddArray(data);
    }

    // Create scalar field (at elements)
    for (auto &p : r.scalars_at_elems)
    {
        vtkDblArrP data = vtkDblArrP::New();
        data->SetNumberOfComponents(1);
        data->SetNumberOfTuples(msh->elems.size());
        data->SetName(p.first.c_str());
        for (size_t i = 0; i < msh->elems.size(); ++i)
            data->SetValue(msh->elems[i]->no - 1, (*(p.second))[i]);
        grid->GetCellData()->AddArray(data);
    }
    // Create vector field (at elems)
    for (auto &p : r.vectors_at_elems)
    {
        vtkDblArrP data = vtkDblArrP::New();
        data->SetNumberOfComponents(3);
        data->SetNumberOfTuples(msh->elems.size());
        data->SetName(p.first.c_str());
        for (size_t i = 0; i < msh->elems.size(); ++i)
            data->SetTuple3(msh->elems[i]->no - 1, (*(p.second))[i](0), (*(p.second))[i](1), (*(p.second))[i](2));
        grid->GetCellData()->AddArray(data);
    }
    // Create tensor field (at elems)
    for (auto &p : r.tensors_at_elems)
    {
        vtkDblArrP data = vtkDblArrP::New();
        data->SetNumberOfComponents(9);
        data->SetNumberOfTuples(msh->elems.size());
        data->SetName(p.first.c_str());
        for (size_t i = 0; i < msh->elems.size(); ++i)
            data->SetTuple9(msh->elems[i]->no - 1, (*(p.second))[i](0, 0), (*(p.second))[i](0, 1), (*(p.second))[i](0, 2),
                            (*(p.second))[i](1, 0), (*(p.second))[i](1, 1), (*(p.second))[i](1, 2),
                            (*(p.second))[i](2, 0), (*(p.second))[i](2, 1), (*(p.second))[i](2, 2));
        grid->GetCellData()->AddArray(data);
    }

    // Write to file
    save(fname);
}

/**
 * @brief Create data structure
 * @authors Adrien Crovato, Romain Boman
 */
void VtkExport::create()
{
    // Create points
    vtkPtsP points = vtkPtsP::New();
    points->SetNumberOfPoints(msh->nodes.size());
    for (auto n : msh->nodes)
        points->InsertPoint(n->no - 1, n->pos(0), n->pos(1), n->pos(2));
    grid->SetPoints(points);

    // Create elements
    for (auto e : msh->elems)
    {
        if (e->type() == ElType::HEX8)
        {
            vtkHexaP cell = vtkHexaP::New();
            for (size_t i = 0; i < e->nodes.size(); ++i)
                cell->GetPointIds()->SetId(i, e->nodes[i]->no - 1);
            grid->InsertNextCell(cell->GetCellType(), cell->GetPointIds());
        }
        else if (e->type() == ElType::TETRA4)
        {
            vtkTetraP cell = vtkTetraP::New();
            for (size_t i = 0; i < e->nodes.size(); ++i)
                cell->GetPointIds()->SetId(i, e->nodes[i]->no - 1);
            grid->InsertNextCell(cell->GetCellType(), cell->GetPointIds());
        }
        else if (e->type() == ElType::QUAD4)
        {
            vtkQuadP cell = vtkQuadP::New();
            for (size_t i = 0; i < e->nodes.size(); ++i)
                cell->GetPointIds()->SetId(i, e->nodes[i]->no - 1);
            grid->InsertNextCell(cell->GetCellType(), cell->GetPointIds());
        }
        else if (e->type() == ElType::TRI3)
        {
            vtkTriP cell = vtkTriP::New();
            for (size_t i = 0; i < e->nodes.size(); ++i)
                cell->GetPointIds()->SetId(i, e->nodes[i]->no - 1);
            grid->InsertNextCell(cell->GetCellType(), cell->GetPointIds());
        }
        else if (e->type() == ElType::LINE2)
        {
            vtkLineP cell = vtkLineP::New();
            for (size_t i = 0; i < e->nodes.size(); ++i)
                cell->GetPointIds()->SetId(i, e->nodes[i]->no - 1);
            grid->InsertNextCell(cell->GetCellType(), cell->GetPointIds());
        }
        else if (e->type() == ElType::POINT1)
        {
            vtkVtxP cell = vtkVtxP::New();
            for (size_t i = 0; i < e->nodes.size(); ++i)
                cell->GetPointIds()->SetId(i, e->nodes[i]->no - 1);
            grid->InsertNextCell(cell->GetCellType(), cell->GetPointIds());
        }
        else
        {
            std::stringstream err;
            err << "tboxVtk::create: translation to VTK for element type "
                << e->type() << " is not supported yet!\n";
            throw std::runtime_error(err.str());
        }
    }

    // Create tags
    vtkIntArrP tag = vtkIntArrP::New();
    tag->SetName("tag");
    tag->SetNumberOfComponents(1);
    tag->SetNumberOfTuples(msh->elems.size());
    for (auto &ptag : msh->ptags)
    {
        for (size_t i = 0; i < ptag.second->elems.size(); ++i)
            tag->SetValue(ptag.second->elems[i]->no - 1, ptag.second->no);
    }
    grid->GetCellData()->AddArray(tag);
}

void VtkExport::write(std::ostream &out) const
{
    out << "VtkExport on " << *msh << std::endl;
}